class Page:
    """
    Creates a Page object that can be converted to amrkdown and pushed to a wiki
    Object must be:
        - a dict
        - a class with a __dict__ attribute
        - a class with a serialize() method
    The default wiki is the built in WIkiJS container. Overwrite with your own wiki API by setting the `wiki_api` class attribute.
    Inherit from autonomous.wiki.Wiki abstract class for the required interface.
    """

    wiki_api = None

    def __init__(self, obj):
        if hasattr(obj, "serialize()"):
            self.obj = obj.serialize()
        elif hasattr(obj, "__dict__"):
            self.obj = obj.__dict__
        elif isinstance(obj, dict):
            self.obj = obj
        else:
            raise TypeError(f"Cannot create Page from {type(obj)}")

    def __str__(self):
        return f"Page(title={self.title}, content={self.content})"

    def __repr__(self):
        return str(self)



class WikiJSPage:
    def __init__(self, **kwargs):
        self.title: str = kwargs.get("title")
        self.description: str = kwargs.get("description")
        self.content: str = kwargs.get("content")
        self.id: int = kwargs.get("id")
        self.path: str = kwargs.get("path")
        self.updatedAt: str = kwargs.get("updatedAt")
        self.tags: list = []

        for t in kwargs.get("tags", []):
            if title := t.get("title"):
                self.tags.append(title)



import json
import uuid

import requests
from authlib.integrations.requests_client import OAuth2Auth, OAuth2Session

from autonomous import log


class AutoAuth:
    def __init__(
        self,
        client_id,
        client_secret,
        issuer,
        redirect_uri,
        scope,
        token_endpoint,
        state=None,
    ):
        """
        Initializes the OpenIDAuth object with the client ID, client secret, and issuer URL.
        """
        self.state = state or uuid.uuid4().hex
        self.client_id = client_id
        self.client_secret = client_secret
        self.issuer = issuer
        self.redirect_uri = redirect_uri
        log(token_endpoint)
        self.token_endpoint = token_endpoint
        self.session = OAuth2Session(
            self.client_id,
            client_secret=self.client_secret,
            scope=scope,
            redirect_uri=self.redirect_uri,
            token_endpoint=self.token_endpoint,
            state=self.state,
        )

    def authenticate(self):
        """
        Initiates the authentication process.
        Returns a redirect URL which should be used to redirect the user to the OpenID provider for authentication.
        """
        uri, state = self.session.create_authorization_url(self.issuer)
        # log(uri, state)
        return uri, state

    def handle_response(self, response, state=None):
        """
        Handles the authentication response from the OpenID provider.
        The response should be a dictionary containing the OpenID provider's response.
        """
        log(self.token_endpoint, response)
        token = self.session.fetch_token(
            authorization_response=response,
            state=state,
        )
        # log(token)

        userinfo = requests.get(self.req_uri, auth=OAuth2Auth(token))
        log(userinfo.text)
        return userinfo.json(), token



import sys

from app.models.model import Model

from autonomous import log


class ChildModel(Model):
    pass


class TestApp:
    def test_model(self):
        m = Model(name="test")
        m.save()
        log(m.pk)
        assert m.pk

    def test_child_model(self):
        m = ChildModel(name="test")
        m.save()
        assert m.table().name == "ChildModel"
        assert m.table().table.name == "ChildModel"



import logging
import os

from config import Config
from flask import Flask

from views.auth import auth_page
from views.admin import admin_page
from views.index import index_page

from autonomous import log
from autonomous.assets import build_assets


def create_app():
    app = Flask(os.getenv("APP_NAME", __name__))
    app.config.from_object(Config)
    # log(app.config.get("PORT"))
    #################################################################
    #                             Filters                           #
    #################################################################
    # app.jinja_env.filters['datetime_format'] = datefilters.datetime_format

    #################################################################
    #                             Extensions                        #
    #################################################################

    # - Assets

    # csspath_dir = "static/style/sass/main.scss"
    # cssoutput_dir = "static/style/main.css"
    # jspath_dir = "static/js"
    # jsoutput_dir = "static/main.min.js"
    app.before_request(lambda: build_assets())

    #################################################################
    #                             ROUTES                            #
    #################################################################

    ######################################
    #           Blueprints               #
    ######################################
    app.register_blueprint(index_page)
    app.register_blueprint(auth_page, url_prefix="/auth")
    app.register_blueprint(admin_page, url_prefix="/admin")

    return app



import time

from autonomous import log
from autonomous.tasks import AutoTasks


def mocktask():
    log("MockTask - starting")
    time.sleep(1)
    log("MockTask - complete")
    return "success"


def longmocktask():
    log("LongTask - Starting")
    time.sleep(30)
    log("LongTask - Complete")
    return "success"


def parametermocktask(*args, **kwargs):
    log("ParameterMockTask - Starting", args, kwargs)
    result = args[0] + args[1]
    log("ParameterMockTask - complete", result)
    return result


def errormocktask():
    log("ErrorMockTask - Starting")
    raise Exception("ErrorMockTask")



# Built-In Modules

# external Modules
from datetime import datetime

from flask import Blueprint, redirect, render_template, request, session, url_for

from autonomous import log
from autonomous.auth import AutoUser, GithubAuth, GoogleAuth

auth_page = Blueprint("auth", __name__)


@auth_page.route("/login", methods=("GET", "POST"))
def login():
    session["user"] = None
    last_login = datetime(session["user"]["last_login"]) == datetime.now()
    if session.get("user"):
        last_login = datetime.fromisoformat(session["user"]["last_login"])
        diff = datetime.now() - last_login
        if diff.days < 30 and session["user"]["state"] == "authenticated":
            return redirect(url_for("index.index"))
        else:
            session["user"] = None
    if request.method == "POST":
        if request.form.get("authprovider") == "google":
            authorizer = GoogleAuth()
            session["authprovider"] = "google"
        elif request.form.get("authprovider") == "github":
            authorizer = GithubAuth()
            session["authprovider"] = "github"
        uri, state = authorizer.authenticate()
        session["authprovider_state"] = state
        # log(uri, state)
        return redirect(uri)
    else:
        return render_template("login.html")


@auth_page.route("/authorize", methods=("GET", "POST"))
def authorize():
    # log(request.args)
    if session["authprovider"] == "google":
        authorizer = GoogleAuth()
    elif session["authprovider"] == "github":
        authorizer = GithubAuth()
    response = str(request.url)
    # log(response)
    user_info, token = authorizer.handle_response(
        response, state=request.args.get("state")
    )
    user = AutoUser.authenticate(user_info, token)
    session["user"] = user.serialize()
    return redirect(url_for("auth.login"))


@auth_page.route("/logout", methods=("POST",))
def logout():
    if session.get("user"):
        user = AutoUser(pk=session["user"]["pk"])
        user.auth["state"] = "unauthenticated"
        user.save()
        session.pop("user")
        # log(f"User {user.name} logged out")
    return redirect(url_for("auth.login"))



import importlib
import os
import subprocess

from redis import Redis
from rq import Queue, Worker
from rq.job import Job

from autonomous import log


class AutoTask:
    def __init__(self, job):
        self.job = job

    @property
    def id(self):
        return self.job.id

    @property
    def status(self):
        status = self.job.get_status()
        if status in ["running", "queued", "started"]:
            return "running"
        return status

    @property
    def running(self):
        return self.job.status == "running"

    @property
    def finished(self):
        result = self.job.status == "finished"
        return result

    @property
    def failed(self):
        result = self.job.status == "failed"
        return result

    def result(self):
        result = self.job.latest_result()
        return result

    @property
    def return_value(self):
        result = self.job.return_value()
        return result


class AutoTasks:
    _connection = None
    queue = None
    workers = []
    all_tasks = []
    config = {
        "host": os.environ.get("REDIS_HOST", ""),
        "port": os.environ.get("REDIS_PORT", ""),
        "password": os.environ.get("REDIS_PASSWORD"),
        "username": os.environ.get("REDIS_USERNAME"),
        "db": os.environ.get("REDIS_DB", 0),
    }

    def __init__(self, queue="default", num_workers=3):
        if not AutoTasks._connection:
            options = {}

            if AutoTasks.config.get("username"):
                options["username"] = AutoTasks.config.get("username")
            if AutoTasks.config.get("username"):
                options["password"] = AutoTasks.config.get("password")
            if AutoTasks.config.get("db"):
                options["db"] = AutoTasks.config.get("db")

            AutoTasks._connection = Redis(
                host=AutoTasks.config.get("host"),
                port=AutoTasks.config.get("port"),
                **options,
            )
        AutoTasks.queue = Queue(queue, connection=AutoTasks._connection)

    def task(self, func, *args, **kwargs):
        """
        :param job: job function
        :param args: job function args
        :param kwargs: job function kwargs
        args and kwargs: use these to explicitly pass arguments and keyword to the underlying job function. This is useful if your function happens to have conflicting argument names with RQ, for example description or ttl.
        :return: job
        """
        job = AutoTasks.queue.enqueue(func, *args, **kwargs)
        self.create_worker(func)
        new_task = AutoTask(job)
        AutoTasks.all_tasks.append(new_task)
        return new_task

    def create_worker(self, func):
        # Get the module containing the target_function
        module = func.__module__

        # Get the file path of the module
        module_path = importlib.import_module(module).__file__

        # Set the PYTHONPATH environment variable
        pythonpath = os.path.dirname(module_path)
        env = os.environ.copy()
        env["PYTHONPATH"] = pythonpath

        rq_user_pass = f"{self.config['username']}:{self.config['password']}"
        rq_url = f"{self.config['host']}:{self.config['port']}"
        rq_db = self.config["db"]
        rq_worker_command = (
            f"rq worker --url redis://{rq_user_pass}@{rq_url}/{rq_db} --burst"
        )

        worker = subprocess.Popen(rq_worker_command, shell=True, env=env)
        self.workers.append(worker)

    # get job given its id
    def get_task(self, job_id):
        # breakpoint()
        task = AutoTasks.queue.fetch_job(job_id)
        return AutoTask(task)

    # get job given its id
    def get_tasks(self):
        return [AutoTask(w) for w in Worker.all(queue=AutoTasks.queue)]

    def clear(self):
        AutoTasks.queue.empty()
        AutoTasks.all_tasks = []


# if __name__ == "__main__":
#     autotasks = AutoTasks()
#     for _ in range(autotasks.workers):
#         create_worker(autotasks.queue.name)



import json
import os
import uuid

from dotenv import load_dotenv

load_dotenv()

import autogen


class AutoTeam:
    config_list = [
        {
            "model": "gpt-4",
            "api_key": os.getenv("OPENAI_API_KEY"),
        },
        {
            "model": "gpt-4-32k",
            "api_key": os.getenv("OPENAI_API_KEY"),
        },
        {
            "model": "gpt-3.5-turbo",
            "api_key": os.environ.get("OPENAI_API_KEY"),
        },
    ]

    def __init__(self, config_path=None):
        with open(config_path, "r") as f:
            aiteam_config = json.load(f)
        self.seed = aiteam_config.get("seed", 42)
        self.max_round = aiteam_config.get("max_round", 20)
        self.temperature = aiteam_config.get("temperature", 0)
        self.input_mode = aiteam_config.get("input_mode", "NEVER")
        self.max_reply = aiteam_config.get("max_reply", 10)

        self.proxy = autogen.UserProxyAgent(
            name="user_proxy",
            human_input_mode=self.input_mode,
            max_consecutive_auto_reply=self.max_reply,
            is_termination_msg=lambda x: x.get("content", "")
            .rstrip()
            .endswith("TERMINATE"),
            code_execution_config={
                "work_dir": "coding",
                "use_docker": False,  # set to True or image name like "python:3" to use docker
            },
        )

        self.agents = {}
        for agent in aiteam_config.get(
            "agents",
            {
                "name": "Assistant",
                "role": "You are a helpful, encouraging, and genial AI Assistant.",
            },
        ):
            self.create_assistant(agent["name"], agent["role"])

    def create_assistant(self, name, system_message):
        assistant = autogen.AssistantAgent(
            name=name,
            system_message=system_message,
            llm_config={
                "seed": self.seed,
                "config_list": self.config_list,
                "temperature": self.temperature,
            },
        )
        self.agents[name] = assistant
        return assistant

    def generate(self, message):
        # the assistant receives a message from the user_proxy, which contains the task description

        groupchat = autogen.GroupChat(
            agents=list(self.agents.values()), messages=[], max_round=self.max_round
        )
        self.manager = autogen.GroupChatManager(
            groupchat=groupchat,
            llm_config={
                "seed": self.seed,
                "config_list": self.config_list,
                "temperature": self.temperature,
            },
        )
        self.proxy.initiate_chat(self.manager, message=message)

    def solution(self, message):
        self.proxy.send(
            "summarize the solution in an easy-to-understand way", self.manager
        )
        # return the last message the proxy received
        return self.proxy.last_message()["content"]
